/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <vector>

# include "Structures.h"
# include "AtomIDs.h"
//# include "StringUtils.h"
# include "IOUtils.h"
# include "PrintPdb.h"
//# include "/home/jouko/LibraryFiles/LinkedList.h"

void assignAtomIDs(vector<AtomStruct> &Atoms)
{
	bool verbose=false;
	string ElementID, AtomName;
	int natom;

	natom=Atoms.size();

	for (int n=0;n<natom;n++)
	{
		Delete(Atoms[n].AtomName, " ");
		ElementID=Atoms[n].AtomName.substr(0,1);
		if (ElementID=="H") Atoms[n].atomid=HYDROGEN;
		else if (ElementID=="C") Atoms[n].atomid=CARBON;
		else if (ElementID=="N") Atoms[n].atomid=NITROGEN;
		else if (ElementID=="O") Atoms[n].atomid=OXYGEN;
		else if (ElementID=="S" && Atoms[n].ResidueName!="SOD") Atoms[n].atomid=SULFUR;
		else if (ElementID=="F") Atoms[n].atomid=IRON;
		else if (ElementID=="P") Atoms[n].atomid=PHOSPHORUS;
		else if (Atoms[n].ResidueName=="SOD" || Atoms[n].ResidueName=="NA0") Atoms[n].atomid=NA_Plus;
		else if (Atoms[n].AtomName=="IP") Atoms[n].atomid=NA_Plus;
		else if (Atoms[n].AtomName=="CLA") Atoms[n].atomid=CL;
                else if (Atoms[n].AtomName=="IM") Atoms[n].atomid=CL;
                else
		{
			if (verbose)
			{
				cout <<"Warning unrecognized atom "<<ElementID<<" will be treated as a hydrogen."<<endl;
				printAtomInfo(Atoms[n]);
			}
                        Atoms[n].atomid=HYDROGEN;
		}
	}
	return;
}

string atomIDToStr(int AtomID)
{
        if (AtomID==HYDROGEN) return "H";
        else if (AtomID==CARBON) return "C";
        else if (AtomID==NITROGEN) return "N";
        else if (AtomID==OXYGEN) return "O";
        else if (AtomID==SULFUR) return "S";
        else if (AtomID==IRON) return "FE";
        else if (AtomID==ExcludedVolume) return "ExcludedVolume";
        else if (AtomID==HydrationShell) return "HydrationShell";
        else if (AtomID==PHOSPHORUS) return "P";
        else if (AtomID==CL) return "CL";
        else if (AtomID==NA_Plus) return "NA";
        else if (AtomID==WaterSphere) return "WaterSphere";
        else if (AtomID==CUBE) return "CUBE";
        else return "UNK";
}

