/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <iostream>
# include <vector>
# include <algorithm>

using namespace std;

# include "IOUtils.h"
# include "ReadDcd.h"
# include "ReadPdb.h"
# include "PrintPdb.h"
# include "AtomUtils.h"
# include "Structures.h"
# include "GetNextFrame.h"

//bool verbose=true;

bool GetDcdFileName(string &DcdFile, string DcdFilePaths)
{
	string line;
	fstream file;
	//if (verbose) 
	OpenFile(DcdFilePaths, file, "DcdFilePaths");
	getline(file, line);
	if (DcdFile=="") DcdFile=line;
	else
	{
		while (true)
		{
			if (file.eof()) break;
			if (line==DcdFile)
			{
				getline(file, line);
				if (file.eof()) return false;
				DcdFile=line;
				file.close();
				return true;
			}
			getline(file, line);
		}
	}
	file.close();
	return false;
}

bool GetNextDcdFrame(ProteinStruct &Protein, int &nthstruct, int &nInDcd, int &nthDcd, string &DCDFile, string DCDFilePaths)
{
	bool IsStructure=true;
	//if (verbose) 
	cout <<"nthDcd= "<<nthDcd<<" nthstruct= "<<nthstruct<<endl;
	if ((nthDcd==1 && nthstruct==1) || DCDFile=="")
	{
		IsStructure=GetLineInFile(DCDFilePaths, DCDFile, nthDcd);
		ReadDcd(DCDFile, Protein, nthstruct, nInDcd);
	}
	if (nthstruct>nInDcd) 
	{
		nthDcd++;
		//cout <<"nthstruct= "<<nthstruct<<" nInDcd= "<<nInDcd<<" nthDcd= "<<nthDcd<<endl;
		IsStructure=GetLineInFile(DCDFilePaths, DCDFile, nthDcd);
		//cout <<"IsStructure= "<<IsStructure<<" DCDFilePaths= "<<DCDFilePaths<<" DCDFile= "<<DCDFile<<endl;
		if (!IsStructure || DCDFile=="") 
                {
                        return false;
                }
		nthstruct=1;
	}	
	ReadDcd(DCDFile, Protein, nthstruct, nInDcd);
	nthstruct++;
	return IsStructure;
}

bool GetNextPdbFrame(ProteinStruct &Protein, string PdbFilePaths, int &nthstruct)
{
	bool IsStructure;
	string PdbFilePath;
	//if (verbose) 
	IsStructure=GetLineInFile(PdbFilePaths, PdbFilePath, nthstruct);
	if (IsStructure) readPdb(PdbFilePath, Protein);
        nthstruct++;
	return IsStructure;
}

bool GetNextFrame(ProteinStruct &Protein, int &nthstruct, int &nInDcd, int &nthDcd, string &DCDFile, string DCDFilePaths, string PdbFilePaths)
{
	bool IsStructure;
	//if (verbose) cout <<"in GetNextFrame. DCDFilePaths= "<<DCDFilePaths<<endl;
        if (DCDFilePaths!="") IsStructure=GetNextDcdFrame(Protein, nthstruct, nInDcd, nthDcd, DCDFile, DCDFilePaths);
	else if (PdbFilePaths!="") IsStructure=GetNextPdbFrame(Protein, PdbFilePaths, nthstruct);
	else
	{
		cout <<"Fatal error. No DCDFiles or PDBFiles given"<<endl;
		exit(EXIT_FAILURE);
	}
	atomAlias(Protein.Atoms);
        assignAtomIDs(Protein.Atoms);
	//getAtomType(Atoms, !params.UniformHydrationShell);
	getAtomType(Protein.Atoms, true);
	getAtomType2(Protein.Atoms);
	return IsStructure;
}
