/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <iostream>
# include <vector>

using namespace std;

# include "MathUtils.h"
# include "VectorManip.h"
# include "MinMax.h"
# include "MonteCarlo.h"
# include "MonteCarloTemplate.h"


Real updateMoveSize(MoveStruct &move, int numToUse, int index)
{
	int nmove=move.dScore.size();
	int count=0;
        Real sum=0, totalChange=0;
        Real acceptProb, change;
	Real moveSizeProb, width;
	Real naccept=0;

        for (int i=nmove-1;i>=0;i--)
        {
		if (move.index[i]==index)
		{
                	acceptProb=move.acceptProb[i];
			moveSizeProb=move.moveSizeProb[i];
                	change=abs(move.dScore[i])*acceptProb/moveSizeProb;
                	sum+=move.size[i]*move.size[i]*change;
			//cout <<"sum= "<<sum<<" size= "<<move.size[i]<<" change= "<<change<<" acceptProb= "<<acceptProb<<" moveSizeProb= "<<moveSizeProb<<endl;
                	totalChange+=change;
			//if (move.dScore[i]!=0) naccept+=acceptProb;
			naccept+=acceptProb;
			if (naccept>Real(numToUse)) break;
			//cout <<"acceptProb= "<<acceptProb<<" moveSizeProb= "
			//<<moveSizeProb<<" change= "<<change<<" move.dScore= "
			//<<move.dScore[i]<<" width= "<<move.width[index]
			//<<" size= "<<move.size[i]<<endl;
		}
        }
	if (count==0) return move.width[index];
	width=sqrt(sum/totalChange);
	if (width==0 || naccept<1.0) width=move.width[index]*0.99;
	if (!isnan(width) && !isinf(width)) return width;
	else 
	{
		cout <<"sum= "<<sum<<" totalChange= "<<totalChange
		<<" index= "<<index<<" width= not a number"<<endl;
		if (isnan(sum) || isnan(totalChange)) 
		{
			sum=0;
			totalChange=0;
			for (int i=nmove-1;i>=0;i--)
			{
                		acceptProb=move.acceptProb[i];
				moveSizeProb=move.moveSizeProb[i];
                		change=abs(move.dScore[i])*acceptProb/moveSizeProb;
                		sum+=move.size[i]*move.size[i]*change;
				cout <<"sum= "<<sum<<" size= "<<move.size[i]<<" change= "<<change<<" acceptProb= "<<acceptProb<<" moveSizeProb= "<<moveSizeProb<<endl;
                		totalChange+=change;
				naccept+=acceptProb;
				if (naccept>Real(numToUse)) break;
				cout <<"acceptProb= "<<acceptProb<<" moveSizeProb= "
				<<moveSizeProb<<" change= "<<change<<" move.dScore= "
				<<move.dScore[i]<<" width= "<<move.width[index]
				<<" size= "<<move.size[i]<<endl;
			}
			string errorStr;
			errorStr="sum= "+toStr(sum)+" totalChange= ";
			errorStr+=toStr(totalChange);
			error(errorStr, __LINE__, __FILE__);
		}
		return move.width[index]*0.99;
	}
}

void updateMoveSizes(MoveStruct &move, int nMoveTypes)
{
	int numToUse=100;
	for (int i=0;i<nMoveTypes;i++)
	{
		move.width[i]=updateMoveSize(move, numToUse, i);
		//cout <<"i= "<<i<<" width= "<<move.width[i]<<" temp= "<<move.temp[i]<<endl;
	}
	//cout <<endl;
}

Real updateMoveProbability(MoveStruct &move, int numToUse, int index)
{
	int nmove=move.dScore.size();
        Real sum=0;
        Real acceptProb, change;
	Real naccept=0;
	Real sumTime=0;
	Real moveProbability;

        for (int i=nmove-1;i>=0;i--)
        {
		if (move.index[i]==index)
		{
                	acceptProb=move.acceptProb[i];
                	change=abs(move.dScore[i])*acceptProb;
                	sum+=change;
			naccept+=acceptProb;
			sumTime+=move.time[i];
			if (naccept>Real(numToUse)) break;
		}
        }
	if (sumTime==0) return 0;
	moveProbability=sum/sumTime;
	if (isnan(moveProbability))
	{
		sum=sumTime=0;
        	for (int i=nmove-1;i>=0;i--)
        	{
			if (move.index[i]==index)
			{
                		acceptProb=move.acceptProb[i];
				cout <<"acceptProb= "<<acceptProb<<endl;
                		change=abs(move.dScore[i])*acceptProb;
				cout <<"change= "<<change<<" dScore= "<<move.dScore[i]<<endl;
                		sum+=change;
				cout <<"sum= "<<sum<<endl;
				naccept+=acceptProb;
				cout <<"naccept= "<<naccept<<endl;
				sumTime+=move.time[i];
				cout <<"sumTime= "<<sumTime<<endl;
				if (naccept>Real(numToUse)) break;
			}
        	}
		error("moveProbability=nan", __LINE__, __FILE__);
	}

	return moveProbability;
}

void normalizeProbabilities(MoveStruct &move)
{
	int ntype=move.probability.size();
	Real sum=0;

	for (int i=0;i<ntype;i++) sum+=move.probability[i];
	for (int i=0;i<ntype;i++) move.probability[i]/=sum;
	move.probSum[0]=0;
	for (int i=1;i<ntype+1;i++) 
	{
		move.probSum[i]=move.probability[i-1]+move.probSum[i-1];
	}
}

void updateMoveProbabilities(MoveStruct &move, int nMoveTypes)
{
	int numToUse=100;
	Real w=0.9;
	Vector oldProbability=move.probability;
	for (int i=0;i<nMoveTypes;i++)
	{
		move.probability[i]=updateMoveProbability(move, numToUse, i);
	}
	normalizeProbabilities(move);
	for (int i=0;i<nMoveTypes;i++)
	{
		move.probability[i]=oldProbability[i]*w+(1.0-w)*move.probability[i];
	}
	normalizeProbabilities(move);
}

void updateTemperatures(Vector &swapRates, Vector &temps, Real targetSwapRate)
{
	//Journal of the Physical Society of Japan Vol. 65, No. 6 June, 1996,
	//pp. 1604-1608
	int ntemp=temps.size();
	Real newTemp, w=0.02;
	Vector oldTemps;

	cout <<"Temperatures: "<<endl;
	oldTemps=temps;
	for (int i=1;i<ntemp;i++)
	{
		newTemp=temps[i-1]+(oldTemps[i]-oldTemps[i-1])*swapRates[i-1]/targetSwapRate;
		temps[i]=oldTemps[i]*(1-w)+newTemp*w;
		sort(temps);
		cout <<"i= "<<i<<" temps= "<<temps[i]<<" swapRate= "<<swapRates[i-1]<<" oldTemp[i]= "<<oldTemps[i]<<" oldTemp[i-1]= "<<oldTemps[i-1]<<" newTemp= "<<newTemp<<endl;
	}
}

Real calcSwapRate(MoveStruct &move1, MoveStruct &move2, int nlast)
{
	int nmove=move1.score.size();
	int ntemp=move1.temp.size();
	Real sum=0;

	if (nmove!=ntemp)
	{
		error("nmove!=ntemp", __LINE__, __FILE__);
	}

	if (nlast>nmove) nlast=nmove;

	for (int i=nmove-1;i>=nmove-nlast;i--)
	{
		sum+=calcSwapProb(move1.score[i], move2.score[i], move1.temp[i], move2.temp[i]);
	}
	return sum/Real(nlast);
}

void calcSwapRates(vector<MoveStruct> &moves, int nlast, Vector &swapRates)
{
	int nrep=moves.size();

	SafeAlloc(swapRates, nrep-1, "swapRates");
	for (int i=0;i<nrep-1;i++)
	{
		swapRates[i]=calcSwapRate(moves[i], moves[i+1], nlast);
	}
}

void updateTemperatures(vector<MoveStruct> &moves, Vector &temps, Real &targetSwapRate, Real targetTemperature)
{
	int nrep=moves.size();
	int nmove, nlast=1000;
	Real maxTemp, w=0.01;
	Vector swapRates;

	nmove=moves[nrep-1].temp.size();
	maxTemp=moves[nrep-1].temp[nmove-1];
	calcSwapRates(moves, nlast, swapRates);
	if (maxTemp>targetTemperature) targetSwapRate=targetSwapRate*(1.0-w)+w;
	else targetSwapRate=2.0*targetSwapRate-targetSwapRate*(1.0-w)-w;	
	cout <<"maxTemp= "<<maxTemp<<" targetTemperature= "<<targetTemperature<<" targetSwapRate= "<<targetSwapRate<<endl;
	updateTemperatures(swapRates, temps, targetSwapRate);
}

void makeRandomChange(Vector &coefficient, MoveStruct &move)
{
	int pick, Size;
	Real moveSize, moveSizeProb;

	Size=coefficient.size();
	pick=randInt(0,Size-1);
	moveSize=randGaussian(move.width[pick]);
	moveSizeProb=calcGaussian(moveSize, move.width[pick]);

	coefficient[pick]+=moveSize;
	SafePushBack(move.index, pick, "index");
	SafePushBack(move.size, moveSize, "size");
	SafePushBack(move.moveSizeProb, moveSizeProb, "moveSizeProb");
}

void makeRandomChange2(Vector &coefficient)
{
	int Size;
	Real max, move;

	Size=coefficient.size();
	
	for (int i=0;i<Size;i++)
	{
		max=coefficient[i]*0.01;
		if (max==0) max=1.0e-10;
		move=randDouble(-max, max);
		coefficient[i]+=move;
		if (coefficient[i]<0) coefficient[i]=-coefficient[i];
	}
}

Real calcAcceptProb(Real dScore, Real temp)
{
	if (dScore>0) return exp(-dScore/temp);
	else return 1.0;
}

Real calcAcceptRate(MoveStruct &move)
{
	return calcAverage(move.acceptProb);
}

Real calcStdDev(MoveStruct &move)
{
	int nstep=move.dScore.size();
	Real dScore, numAccept, sum, sumSqr, ave, aveSqr, sqrAve, stdDev;

	numAccept=0;
	sum=0;
	sumSqr=0;
	stdDev=0;

	for (int i=0;i<nstep;i++)
	{
		dScore=move.dScore[i];
		numAccept+=move.acceptProb[i];
		sum+=dScore*move.acceptProb[i];
		sumSqr+=dScore*dScore*move.acceptProb[i];
	}
	if (numAccept>0)
	{
		ave=sum/numAccept;
		aveSqr=sumSqr/numAccept;
		sqrAve=ave*ave;
		stdDev=sqrt(aveSqr-sqrAve);
	}
	return stdDev;
}

void setTemps(Vector &temps, Real minTemp, Real maxTemp, int nrep)
{
	Real scale;
	//Real c=2.6;
	Real c=1.0;
	SafeAlloc(temps, nrep, "temps");
	scale=power(maxTemp/minTemp, 1.0/Real(nrep-1));
	temps[0]=minTemp;
	for (int i=1;i<nrep;i++)
	{
		//temps[i]=temps[i-1]*scale;
		scale=power(Real(i)/Real(nrep-1), c);
		temps[i]=minTemp*power(maxTemp/minTemp, scale);
	}
}

void initializeCoefficients(Vector &coefficient, Matrix &coefficients, int nrep)
{
	int Size1, Size2;
	coefficients.clear();
	for (int i=0;i<nrep;i++)
	{
		SafePushBack(coefficients, coefficient, "coefficients");
	}
	Get2DVectorSize(coefficients, Size1, Size2);
	for (int i=0;i<Size1;i++)
	{
		for (int j=0;j<Size2;j++)
		{
			coefficients[i][j]*=randDouble(-1.0, 1.0);
		}
	}
}

void initializeMoveProbabilities(MoveStruct &move, int nMoveTypes)
{
	SafeAlloc(move.probability, nMoveTypes, "probability");
	SafeAlloc(move.probSum, nMoveTypes+1, "probSum");

	for (int i=0;i<nMoveTypes;i++)
	{
		move.probability[i]=1.0/Real(nMoveTypes);
	}
	normalizeProbabilities(move);
}

void initializeMove(Vector &coefficient, MoveStruct &move)
{
	int Size=coefficient.size();
	SafeAlloc(move.width, Size, "width");
	for (int i=0;i<Size;i++)
	{
		move.width[i]=abs(coefficient[i]*0.1);
	}
	initializeMoveProbabilities(move, Size);
}

void initializeMoves(Matrix &coefficients, vector<MoveStruct> &moves)
{
	int nrep=coefficients.size();
	MoveStruct tempMove;
	SafeAlloc(moves, tempMove, nrep, "moves");
	for (int i=0;i<nrep;i++)
	{
		initializeMove(coefficients[i], moves[i]);
	}
}

Real calcSwapProb(Real score1, Real score2, Real temp1, Real temp2)
{
	Real preExp, prob;

	preExp=(score1-score2)*(1.0/temp1-1.0/temp2);
	if (preExp>=0) prob=1.0;
	else if (preExp<0 && preExp>-20.0) prob=exp(preExp);
	else prob=0;

	return prob;
}

Real calcAveSwapProb(MoveStruct &move1, MoveStruct &move2, Real temp1, Real temp2)
{
	int nmove1, nmove2, Size;
	Real sum=0;

	nmove1=move1.score.size();
	nmove2=move2.score.size();

	Size=min(nmove1, nmove2);

	temp1=move1.temp[0];
	temp2=move2.temp[0];
	for (int i=0;i<Size;i++)
	{
		sum+=calcSwapProb(move1.scoreAccept[i], move2.scoreAccept[i], temp1, temp2);
	}
	return sum/Real(Size);
}

bool swapTemps(Vector &temps, Vector &score, Matrix &coefficients, Vector &totalSwaps, Vector &swapAttempts, vector<int> &hotOrCold, int cycle)
{
	bool boolSwap=false;
	int nreplicas=temps.size();
	int Size;
	Real prob, rand;

	for (int i=cycle%2;i<nreplicas-1;i+=2)
	{
		prob=calcSwapProb(score[i], score[i+1], temps[i], temps[i+1]);
		totalSwaps[i]+=prob; swapAttempts[i]+=1.0;
		//cout <<"score1= "<<score[i]<<" score2= "<<score[i+1]
		//<<" temp1= "<<temps[i]<<" temp2= "<<temps[i+1]<<endl;
		//cout <<"prob= "<<prob<<" dScore= "<<score[i]-score[i+1]
		//<<" dTemp= "<<1.0/temps[i]-1.0/temps[i+1]<<endl;
		rand=randDouble(0, 1.0);
		if (rand<prob)
                {
                        swap(score[i], score[i+1]);
			swap(hotOrCold[i], hotOrCold[i+1]);
			Size=coefficients[i].size();
			for (int j=0;j<Size;j++)
			{
                        	swap(coefficients[i][j], coefficients[i+1][j]);
			}
			boolSwap=true;	
                }
	}
	hotOrCold[0]=COLD;
	hotOrCold[nreplicas-1]=HOT;
	return boolSwap;
}

void incrementHotColdHistograms(vector<int> &hotOrCold, Vector &hotHistogram, Vector &coldHistogram)
{
	int nrep=hotOrCold.size();

	for (int i=0;i<nrep;i++)
	{
		if (hotOrCold[i]==HOT) hotHistogram[i]+=1.0;
		else if (hotOrCold[i]==COLD) coldHistogram[i]+=1.0;
	}
}

void fixFractionHot(Vector &fractionHot, int index)
{
	int nrep=fractionHot.size();
	int next=nrep;
	Real slope;

	for (int i=index+1;i<nrep;i++)
	{
		if (fractionHot[i]>fractionHot[index])
		{
			next=i;
			break;
		}
	}
	slope=(fractionHot[next]-fractionHot[index])/Real(next-index);
	for (int i=index+1;i<next;i++)
	{
		fractionHot[i]=Real(i-index)*slope+fractionHot[index];
	}
}

void fixFractionHot(Vector &fractionHot)
{
	int nrep=fractionHot.size();

	for (int i=0;i<nrep;i++)
	{
		if (fractionHot[i+1]<fractionHot[i])
		{
			fixFractionHot(fractionHot, i);
		}
	}
	cout <<"i\tfractionHot\ttarget"<<endl;
	for (int i=0;i<nrep;i++)
	{
		cout <<i<<"\t"<<fractionHot[i]<<"\t"<<Real(i)/Real(nrep-1)<<endl;
	}
}

void calcFractionHot(Vector &hotHistogram, Vector &coldHistogram, Vector &fractionHot)
{
	int nhot=hotHistogram.size();
	int ncold=coldHistogram.size();
	int nfrac=fractionHot.size();
	Real total;

	if (nhot!=ncold || nhot==0)
	{
		string errorStr="nhot= "+toStr(nhot);
		errorStr+="ncold= "+toStr(ncold);
		errorStr+="nfrac= "+toStr(nfrac);
		error(errorStr, __LINE__, __FILE__);
	}

	if (nhot!=nfrac) SafeAlloc(fractionHot, nhot, "fractionHot");

	cout <<"i\thot\tcold\tfrac"<<endl;
	for (int i=0;i<nhot;i++)
	{
		total=coldHistogram[i]+hotHistogram[i];
		if (total!=0) fractionHot[i]=hotHistogram[i]/(total);
		else fractionHot[i]=0;
		cout <<i<<"\t"<<hotHistogram[i]<<"\t"<<coldHistogram[i]<<"\t"<<fractionHot[i]<<endl;
	}
}

void calcFractionCold(Vector &hotHistogram, Vector &coldHistogram, Vector &fractionCold)
{
	int nhot=hotHistogram.size();
	int ncold=coldHistogram.size();
	int nfrac=fractionCold.size();
	Real total;

	if (nhot!=ncold || nhot==0)
	{
		string errorStr="nhot= "+toStr(nhot);
		errorStr+="ncold= "+toStr(ncold);
		errorStr+="nfrac= "+toStr(nfrac);
		error(errorStr, __LINE__, __FILE__);
	}

	if (nhot!=nfrac) SafeAlloc(fractionCold, nhot, "fractionCold");

	cout <<"i\thot\tcold\tfrac"<<endl;
	for (int i=0;i<nhot;i++)
	{
		total=coldHistogram[i]+hotHistogram[i];
		if (total!=0) fractionCold[i]=coldHistogram[i]/(total);
		else fractionCold[i]=0;
		cout <<i<<"\t"<<hotHistogram[i]<<"\t"<<coldHistogram[i]<<"\t"<<fractionCold[i]<<endl;
	}
}

void updateTemperatureUsingHotHistogram(Vector &fractionHot, Vector &temps)
{
	int nrep=temps.size();
	Real diff, slope, targetFrac;
	Real w=0.1;
	Vector newTemp;

	SafeAlloc(newTemp, nrep, "newTemp");

	for (int i=1;i<nrep-1;i++)
	{
		targetFrac=Real(i)/Real(nrep-1);
		for (int j=0;j<nrep-1;j++)
		{
			if (fractionHot[j]<targetFrac && fractionHot[j+1]>targetFrac)
			{
				slope=(temps[j+1]-temps[j])/(fractionHot[j+1]-fractionHot[j]);
				diff=targetFrac-fractionHot[j];
				newTemp[i]=temps[j]+slope*diff;
				//cout <<"i= "<<i<<" newTemp= "<<newTemp[i]<<" temps= "<<temps[i]<<" slope= "<<slope<<" diff= "<<diff
				//<<" fractionHot[j]= "<<fractionHot[j]<<" fractionHot[j+1]= "<<fractionHot[j+1]
				//<<" temps[i]= "<<temps[i]<<" temps[i+1]= "<<temps[i+1]<<" targetFrac= "<<targetFrac<<endl;
			}
		}
	}
	for (int i=1;i<nrep-1;i++)
	{
		temps[i]=newTemp[i]*w+(1.0-w)*temps[i];
	}
}

void initializeHotAndCold(vector<int> &hotOrCold, Vector &hotHistogram, Vector &coldHistogram, int nrep)
{
	SafeAlloc(hotOrCold, NEITHER, nrep, "hotOrCold");
	SafeAlloc(hotHistogram, nrep, "hotHistogram");
	SafeAlloc(coldHistogram, nrep, "coldHistogram");
	hotOrCold[0]=COLD;
	hotOrCold[nrep-1]=HOT;
}

void multipleSwaps(Matrix &coefficients, Vector &scores, Vector &totalSwaps, Vector &swapAttempts, Vector &temps, vector<int> &hotOrCold, Vector &hotHistogram, Vector &coldHistogram, int iterations)
{
	bool swapped;

	for (int i=0;i<iterations;i++)
	{
		swapped=swapTemps(temps, scores, coefficients, totalSwaps, swapAttempts, hotOrCold, i);
		incrementHotColdHistograms(hotOrCold, hotHistogram, coldHistogram);
		if (!swapped) break;
	}
}

void multipleSwaps(Matrix &coefficients, Vector &scores, Vector &totalSwaps, Vector &swapAttempts, Vector &temps, vector<int> &hotOrCold, Vector &hotHistogram, Vector &coldHistogram)
{
	int iterations=100;

	multipleSwaps(coefficients, scores, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram, iterations);
}

void calcSwapRates(Vector &totalSwaps, Vector &swapAttempts, Vector &swapRates)
{
	int nrep=totalSwaps.size();
	
	SafeAlloc(swapRates, nrep-1, "swapRates");
	for (int i=0;i<nrep-1;i++)
	{
		swapRates[i]=totalSwaps[i]/swapAttempts[i];
		if (swapAttempts[i]!=0) swapRates[i]=totalSwaps[i]/swapAttempts[i];
		else swapRates[i]=0;
	}
}

void printSwapRates(Vector &totalSwaps, Vector &swapAttempts)
{
	int nrep=totalSwaps.size();
	Real swapRate;

	for (int i=0;i<nrep-1;i++)
	{
		if (swapAttempts[i]!=0) swapRate=totalSwaps[i]/swapAttempts[i];
		else swapRate=0;
		cout <<"swapRate= "<<swapRate<<endl;
	}
}

void printMoveSizes(MoveStruct &move)
{
	printVector(move.width, "width");
}

void printMoveSizes(vector<MoveStruct> &moves)
{
	int nmove=moves.size();

	for (int i=0;i<nmove;i++)
	{
		cout <<endl;
		cout <<"Replica "<<i<<endl;
		printMoveSizes(moves[i]);
	}
	cout <<endl;
}

void printMoveProbabilities(MoveStruct &move)
{
	printVector(move.probability, "probability");
}

void printMoveProbabilities(vector<MoveStruct> &moves)
{
	int nmove=moves.size();

	for (int i=0;i<nmove;i++)
	{
		cout <<endl;
		cout <<"Replica "<<i<<endl;
		printMoveProbabilities(moves[i]);
	}
	cout <<endl;
}
